Attribute VB_Name = "modEvaluation"
'Evaluation Module
'
'Contains all functions and subs for evaluation of a string, including a true/false
'analyzer.  For RPN, see modRPN.

Option Explicit

Public Function GetValue(ByRef Start As Long, ByVal Parameter As String, Optional Eval As Boolean = True, Optional ByVal InWhat As String) As String
    'This will search for the paramter, starting at the specified location, skipping
    'things in quotes, then evalutate it.
    Dim FinalCode As String, WhereItIs As Long
    
    If InWhat = "" Then InWhat = dScript
    WhereItIs = CodeInStr(Start, InWhat, Parameter)
    
    If WhereItIs = 0 Then
        FlagError "Syntax Error", "Expected '" & Parameter & "'", iLocation: Exit Function
    Else
        FinalCode = Mid(InWhat, Start, WhereItIs - Start)
        
        GetValue = Trim(FinalCode)
        If Eval Then GetValue = Solve(GetValue)
        
        Start = Start + Len(FinalCode) + Len(Parameter)
    End If
End Function

Public Function GetValue3(ByRef Start As Long, ByVal InWhat As String) As String
    'Similar to GetValue, however accounts for parenthesis
    Dim FinalCode As String
    Dim MakeInt As Integer, sChar As String
    Dim Depth As Integer, Quote As String
    Dim x As Integer
    
    If InWhat = "" Then InWhat = dScript
    
    For x = Start To Len(InWhat)
        sChar = Mid(InWhat, x, 1)
        If sChar = ")" Then
            If Depth = 0 Then
                FinalCode = Mid(InWhat, Start, x - Start)
                GetValue3 = Trim(FinalCode)
                Start = Start + Len(FinalCode) + 1
                Exit Function
            Else
                Depth = Depth - 1
            End If
        ElseIf sChar = Chr(34) Then
            Quote = InStr(x + 1, InWhat, Chr(34))
            If Quote <> 0 Then
                x = Quote
            End If
        ElseIf sChar = "(" Then
            Depth = Depth + 1
        End If
    Next 'x
    
    FlagError "Syntax Error", "Expected ')'", iLocation: Exit Function
End Function

Public Function GetFunctions(ByVal sFunction As String, Optional flagTypeError As Boolean = True) As typeVar
    'On Error GoTo error
    Dim tArgs() As String, Args() As typeVar
    Dim ProcFunction As String, ProcArguments As String
    Dim tempInt As Integer
    Dim tempVar() As String
    Dim n As Integer, i As Long
    Dim tempString As String
    
    sFunction = Trim(sFunction)
    i = InStr(1, sFunction, "(")
    ProcFunction = Trim(LCase(Mid(sFunction, 1, i - 1)))
    ProcArguments = Trim(GetValue3(i + 1, Trim(sFunction)))

    If ProcArguments <> "" Then 'solve out all given arguments
        tArgs = dSplit(ProcArguments, ",")    'split arguments
        ReDim Args(UBound(tArgs))
        For n = LBound(Args) To UBound(Args) 'solve all arguments
            Args(n) = makeIntoVar(Trim(tArgs(n)))
        Next 'n
    End If
    
    GetFunctions.isArray = False
    GetFunctions.cElement = -1
    
    Select Case ProcFunction
        Case "lcase"
            If UBound(Args) <> 0 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                GetFunctions.value = LCase(Args(0).value)
            End If
        Case "ucase"
            If UBound(Args) <> 0 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                GetFunctions.value = UCase(Args(0).value)
            End If
        Case "trim"
            If UBound(Args) <> 0 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                GetFunctions.value = Trim(Args(0).value)
            End If
        Case "isnumeric"
            If UBound(Args) <> 0 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                GetFunctions.value = IsNumeric(Args(0).value)
            End If
        Case "isop"
            If UBound(Args) <> 0 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                GetFunctions.value = IsOperator(Args(0).value)
            End If
        Case "int"
            If UBound(Args) <> 0 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                GetFunctions.value = Int(Args(0).value)
            End If
        Case "replace"
            If UBound(Args) <> 2 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                GetFunctions.value = Replace(Args(0).value, Args(1).value, Args(2).value)
            End If
        Case "fileexist"
            If UBound(Args) <> 0 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                GetFunctions.value = FileExist(Args(0).value)
            End If
        Case "rndnum"
            If UBound(Args) <> 0 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                Randomize
                GetFunctions.value = Int(Rnd * Int(Args(0).value))
            End If
        Case "rnd"
            If UBound(Args) <> 0 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                Randomize
                GetFunctions.value = (Rnd * Args(0).value)
            End If
        Case "asc"
            If UBound(Args) <> 0 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                GetFunctions.value = Asc(Args(0).value)
            End If
        Case "chr"
            If UBound(Args) <> 0 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                GetFunctions.value = Chr(Args(0).value)
            End If
        Case "hex"
            If UBound(Args) <> 0 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                GetFunctions.value = Hex(Args(0).value)
            End If
        Case "len"
            If UBound(Args) <> 0 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                GetFunctions.value = Len(Args(0).value)
            End If
        Case "val"
            If UBound(Args) <> 0 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                GetFunctions.value = Val(Args(0).value)
            End If
        Case "string"
            If UBound(Args) <> 1 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                GetFunctions.value = String(Args(0).value, Args(1).value)
            End If
        Case "right"
            If UBound(Args) <> 1 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                GetFunctions.value = Right(Args(0).value, Args(1).value)
            End If
        Case "left"
            If UBound(Args) <> 1 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                GetFunctions.value = Left(Args(0).value, Args(1).value)
            End If
        Case "instr"
            If UBound(Args) <> 2 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                GetFunctions.value = InStr(Args(0).value, Args(1).value, Args(2).value)
            End If
        Case "mid"
            If UBound(Args) = 1 Then
                GetFunctions.value = Mid(Args(0).value, Args(1).value)
            ElseIf UBound(Args) = 2 Then
                GetFunctions.value = Mid(Args(0).value, Args(1).value, Args(2).value)
            Else
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            End If
        Case "rpn"
            If UBound(Args) <> 0 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                GetFunctions.value = RPNcalculate(Args(0).value)
            End If
        Case "inarray"
            If UBound(Args) <> 1 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                If Not Args(0).isArray Or Args(0).numElements < 0 Then FlagError "Run-Time Error", "Argument 0 for 'inarray' must be a non-empty array-type variable", iLocation: Exit Function
                GetFunctions.value = 0
                For i = 0 To UBound(Args(0).arrayValues)
                    If LCase(Args(1).value) = LCase(Args(0).arrayValues(i)) Then
                        GetFunctions.value = 1
                        Exit For
                    End If
                Next i
            End If
        Case "find_element"
            If UBound(Args) <> 1 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                If Not Args(0).isArray Or Args(0).numElements < 0 Then FlagError "Run-Time Error", "Argument 0 for 'inarray' must be a non-empty array-type variable", iLocation: Exit Function
                GetFunctions.value = -1
                For i = 0 To UBound(Args(0).arrayValues)
                    If LCase(Args(1).value) = LCase(Args(0).arrayValues(i)) Then
                        GetFunctions.value = i
                        Exit For
                    End If
                Next i
            End If
        Case "rnd_element"
            If UBound(Args) <> 0 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                If Not Args(0).isArray Or Args(0).numElements < 0 Then FlagError "Run-Time Error", "Argument 0 for 'rndelement' must be a non-empty array-type variable", iLocation: Exit Function
                GetFunctions.value = Int(Rnd * (Args(0).numElements + 1))
            End If
        'FILE FUNCTIONS
        'NOTE TO SELF: ADD F_WRITE AND F_READ
        '(if you can read this and you're not azure then yell at him!!)
        Case "readfile"
            If UBound(Args) <> 0 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                GetFunctions.value = ReadFile(Args(0).value)
            End If
        Case "f_open"
            If UBound(Args) <> 1 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                GetFunctions.value = doFileOpen(Args(0).value, Args(1).value)
            End If
        Case "f_out"
            If UBound(Args) <> 1 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                Print #Args(0).value, Args(1).value
            End If
        Case "f_close"
            If UBound(Args) <> 0 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                Close #Args(0).value
            End If
        Case "f_linein"
            If UBound(Args) <> 0 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                Line Input #Args(0).value, tempString
                GetFunctions.value = tempString
            End If
        Case "f_in"
            If UBound(Args) <> 0 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                Input #Args(0).value, Args(1).value
                GetFunctions.value = Args(1).value
            End If
        Case "f_getchunk"
            If UBound(Args) <> 1 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                tempString = Space(Val(Args(1).value))
                Get #Args(0).value, , tempString
                GetFunctions.value = tempString
            End If
        Case "eof"
            If UBound(Args) <> 0 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                GetFunctions.value = EOF(Args(0).value) 'note: accepts file number
            End If
        Case "lof"
            If UBound(Args) <> 0 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                GetFunctions.value = LOF(Args(0).value) 'note: accepts file number
            End If
        Case "loc"
            If UBound(Args) <> 0 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                GetFunctions.value = Loc(Args(0).value) 'note: accepts file number
            End If
        Case "filelen"
            If UBound(Args) <> 0 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                GetFunctions.value = FileLen(Args(0).value) 'note: accepts full file name
            End If
        
        Case "makedir"
            If UBound(Args) <> 0 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                MkDir Args(0).value
            End If
        Case "direxist"
            If UBound(Args) <> 0 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                GetFunctions.value = DirExist(Args(0).value)
            End If
        Case "copyfile"
            If UBound(Args) <> 1 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                FileCopy Args(0).value, Args(1).value
            End If
        Case "rename"
            If UBound(Args) <> 1 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                Name Args(0).value As Args(1).value
            End If
            
        Case "isarray"
            If UBound(Args) <> 0 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                tempInt = GetVarIndex(Args(0).name, True)
                
                GetFunctions.value = dVars(tempInt).isArray
            End If
        Case "isempty"
            If UBound(Args) <> 0 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                tempInt = GetVarIndex(Args(0).name, True)
                
                If dVars(tempInt).isArray Then
                    GetFunctions.value = IIf(dVars(tempInt).numElements = -1, True, False)
                Else
                    GetFunctions.value = IIf(dVars(tempInt).value = "", True, False)
                End If
            End If
        Case "ubound"
            If UBound(Args) <> 0 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                tempInt = GetVarIndex(Args(0).name, True)
                If Not dVars(tempInt).isArray Then FlagError "Run-Time Error", "Variable type mismatch: " & vbCrLf & dVars(tempInt).name & ": darray <-> dvar", iLocation
                
                GetFunctions.value = dVars(tempInt).numElements
            End If
        Case "join"
            If UBound(Args) <> 1 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                If Not Args(0).isArray Or Args(0).numElements < 0 Then FlagError "Run-Time Error", "Argument 0 for 'inarray' must be a non-empty array-type variable", iLocation: Exit Function
                GetFunctions.value = Join(Args(0).arrayValues, Args(1).value)
            End If
            
        'ARRAY FUNCTIONS
        Case "split"
            If UBound(Args) = 1 Then 'Check for correct arguments
                GetFunctions.arrayValues = Split(Args(0).value, Args(1).value)
            ElseIf UBound(Args) = 2 Then
                GetFunctions.arrayValues = Split(Args(0).value, Args(1).value, Args(2).value)
            Else
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            End If
            GetFunctions.isArray = True
            GetFunctions.numElements = UBound(GetFunctions.arrayValues)
        Case "array"    'create an array
            If ProcArguments = "" Then ReDim tempVar(0): GoTo nowhere: Exit Function
            ReDim tempVar(UBound(Args))
            
            For n = 0 To UBound(Args)
                If Args(n).isArray Then
                    FlagError "Run-Time Error", "Type mismatch: " & vbCrLf & Args(n).name & ": dvar <-> darray", iLocation
                End If
                tempVar(n) = Args(n).value
            Next n
nowhere:
            GetFunctions.arrayValues = tempVar
            GetFunctions.isArray = True
            GetFunctions.numElements = UBound(GetFunctions.arrayValues)
            
        'CONSOLE FUNCTIONS
        Case "console.length", "conlen"
            If ProcArguments <> "" Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                If appDOS Then Exit Function
                GetFunctions.value = Len(frmRun.txtText.Text)
            End If
        Case "console.setstart", "setstart"
            If UBound(Args) <> 0 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                If appDOS Then Exit Function
                frmRun.txtText.SelStart = Val(Args(0).value)
            End If
        Case "console.settext", "setconsole"
            If UBound(Args) <> 0 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                If appDOS Then Exit Function
                frmRun.txtText.Text = Args(0).value
            End If
        Case "console.gettext", "getconsole"
            If ProcArguments <> "" Then
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                If appDOS Then Exit Function
                GetFunctions.value = frmRun.txtText.Text
            End If
        Case "debug"
            If UBound(Args) <> 0 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                If appDOS Then Exit Function
                Select Case Args(0).value
                    Case 0
                        frmRun.lblDebug.Visible = False
                        debugMode = False
                    Case 1
                        frmRun.lblDebug.Visible = True
                        debugMode = True
                End Select
            End If
        Case "console.lock"
            If UBound(Args) <> 0 Then 'Check for correct arguments
                FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
            Else
                If appDOS Then Exit Function
                Select Case Args(0).value
                    Case 0
                        frmRun.txtText.Locked = False
                    Case 1
                        frmRun.txtText.Locked = True
                End Select
            End If
        Case Else 'user defined function

            If LCase(ProcFunction) = "main" Then FlagError "Syntax Error", "Cannot call 'main'", iLocation: Exit Function
            
            If Not hasFunctions Then _
                FlagError "Syntax Error", "Function undefined: " & ProcFunction, iLocation: Exit Function
            
            tempInt = GetProcIndex(ProcFunction, True)
            
            Dim NewCode As String
            Dim OldCode As String, OldProc As String, OldPID As Integer
            Dim Args2 As Variant
            Dim returnVarIndex As Integer
            
            NewCode = dProcs(tempInt).Code  'get the function's code from code array
            
            Args2 = dSplit(dProcs(tempInt).Parameters, ",") 'split declaration arguments
            
            'make it so we can return values
            returnCount = returnCount + 1
            ReDim Preserve ProcReturns(returnCount)
            ProcReturns(returnCount).name = ProcFunction
            ProcReturns(returnCount).isArray = dProcs(tempInt).isArray
            
            OldCode = dScript           'save the old code
            OldProc = CurrentProc       'save the old proc
            OldPID = CurrentPID
            pidCount = pidCount + 1
            CurrentPID = pidCount
            CurrentProc = ProcFunction  'set the current proc
            
            If Trim(ProcArguments) <> "" And Trim(dProcs(tempInt).Parameters) <> "" Then
                'check for correct number of args by comparing declaration and provided vars
                If UBound(Args) <> UBound(Args2) Then
                    FlagError "Syntax Error", "Wrong number of arguments for " & ProcFunction & "().", iLocation: Exit Function
                End If
    
                'create vars for this function
                For n = LBound(Args) To UBound(Args)
                    tempInt = AddVar(Trim(Args2(n)), Args(n).value, ProcFunction, CurrentPID, False)
                    If Args(n).isArray Then
                        'FlagError "Syntax Error", "Can't predefine number of elements when declaring argument", iLocation: Exit Function
                        dVars(tempInt).arrayValues = Args(n).arrayValues
                        dVars(tempInt).numElements = Args(n).numElements
                        dVars(tempInt).isArray = True
                    End If
                Next 'n
            End If
            
            'parse all new code, based on this function
            frmRun.LinkCode NewCode, False
            ClearVars CurrentPID  'clear all vars for this function
            ClearProcLoops CurrentPID     'clear old for loops
            CurrentPID = OldPID
            CurrentProc = OldProc   'reset to old proc
            dScript = OldCode       'reset to old code

            'add the return value
            GetFunctions = ProcReturns(returnCount)
            ProcReturns(returnCount).value = ""   'reset returns
            returnCount = returnCount - 1   'ditto
            
    End Select
    
    If GetFunctions.value = "True" Then GetFunctions.value = 1
    If GetFunctions.value = "False" Then GetFunctions.value = 0

    Exit Function
    
error:
    FlagError "Run-Time Error", "Error: " & err.Description, iLocation
End Function

Public Function Solve(ByVal sFunction As String) As String
    'this is the master solver
    
    If Trim(sFunction) = "" Then Exit Function  'makes sure it's something
    
    'sFunction = GetFunctions(sFunction)         'first solve all functions
    Solve = RPNcalculate(sFunction)             'send it to the RPN engine
End Function

Public Function IsOperator(ByVal NumVal As String) As Boolean
    'Determines if a character is an operator
    Select Case NumVal
    Case "+", "-", "*", "/", "\", "&", ">", "<", "=", "^", "%"
        IsOperator = True
    End Select
End Function

Public Function doEval(ByVal LeftVal As String, ByVal Operator As String, ByVal RightVal As String) As Integer
    'Determines if an expression is True or False
    doEval = 0
    Select Case Operator
        Case ">"
            If IsNumeric(LeftVal) Or IsNumeric(RightVal) Then
                If Val(LeftVal) > Val(RightVal) Then doEval = 1
            Else
                If LeftVal > RightVal Then doEval = 1
            End If
        Case "<"
            If IsNumeric(LeftVal) Or IsNumeric(RightVal) Then
                If Val(LeftVal) < Val(RightVal) Then doEval = 1
            Else
                If LeftVal < RightVal Then doEval = 1
            End If
        Case "=", "=="
            If IsNumeric(LeftVal) Or IsNumeric(RightVal) Then
                If Val(LeftVal) = Val(RightVal) Then doEval = 1
            Else
                If LeftVal = RightVal Then doEval = 1
            End If
        Case "<>", "!="
            If IsNumeric(LeftVal) Or IsNumeric(RightVal) Then
                If Val(LeftVal) <> Val(RightVal) Then doEval = 1
            Else
                If LeftVal <> RightVal Then doEval = 1
            End If
        Case ">="
            If IsNumeric(LeftVal) Or IsNumeric(RightVal) Then
                If Val(LeftVal) >= Val(RightVal) Then doEval = 1
            Else
                If LeftVal >= RightVal Then doEval = 1
            End If
        Case "<="
            If IsNumeric(LeftVal) Or IsNumeric(RightVal) Then
                If Val(LeftVal) <= Val(RightVal) Then doEval = 1
            Else
                If LeftVal <= RightVal Then doEval = 1
            End If
        Case Else
            doEval = 0
    End Select
End Function

Public Function Eval(ByVal sFunction As String) As Boolean
    'This parses a string into a left value, operator, and right value and determines
    'if the statement is true or false
    
    Dim LeftVal As String, RightVal As String, Operator
    Dim sChar As String, OpFound As Boolean
    Dim x As Integer
    
    For x = 1 To Len(sFunction)
        sChar = Mid(sFunction, x, 1)
        If sChar = ">" Or sChar = "<" Or sChar = "=" Or sChar = "!" And Not OpFound Then
            Operator = Operator & sChar
            OpFound = True
        Else
            If OpFound Then
                RightVal = RightVal & sChar
            Else
                LeftVal = LeftVal & sChar
            End If
        End If
    Next x
    
    LeftVal = Solve(LeftVal)
    RightVal = Solve(RightVal)
    
    Eval = doEval(LeftVal, Operator, RightVal)
End Function

Public Function FindIdentifier(StartPos As Long, InWhat As String) As String
    On Error Resume Next
    Dim i As Long
    
    For i = StartPos To 1 Step -1
        Select Case Mid(InWhat, i - 1, 1)
            Case " ", ";", vbTab, "("
                If i > StartPos Then GoTo SkipCurrentI
                FindIdentifier = Trim(Mid(InWhat, i, StartPos - (i - 1)))
                If FindIdentifier <> "" Then Exit Function
        End Select
        If i = 2 Then Exit For
        
        If Mid(InWhat, i - 2, 2) = vbCrLf Then
            If i > StartPos Then GoTo SkipCurrentI
            FindIdentifier = Trim(Mid(InWhat, i, StartPos - (i - 1)))
            If FindIdentifier <> "" Then Exit Function
            Exit Function
        End If
SkipCurrentI:
    Next i
    FindIdentifier = Trim(Mid(InWhat, 1, StartPos))
End Function

Public Function isMathThingy(Thingy As String) As Boolean
    isMathThingy = False
    
    If IsOperator(Thingy) Then isMathThingy = True
    If IsNumeric(Thingy) Then isMathThingy = True

End Function
