Attribute VB_Name = "modRPN"
Option Explicit

'This is the RPN (Reverse-Polish Notation) engine.  The RPN is an extremely fast
'and stable way to solve expressions.  Moving D++ over to RPN has increased D++'s
'capabilities dramatically.
'Normally an RPN engine is just used for math functions, however this one has been
'super-modified to accomodate for strings, variables, and functions.


'stuff used for independent testing of the RPN engine
'Sub main()
'
'End Sub

Public Function RPNcalculate(expression As String, Optional equalityStatement As Boolean = False) As String
    'On Error GoTo err
    'main solver
    'IMPORTANT: if equalityStatement is TRUE, then the equal sign (=) operator functions as EQAULITY (i.e. 3=3 will return true)
    '           if equalityStatement is FALSE, then the equal sign (=) operator functions as ASSIGNMENT (i.e. n=3 will assign 3 to n)
    
    Dim aDebug As Boolean
    If Trim(expression) = "" Then Exit Function
    If debugMode Then
        If Left(expression, 1) = "@" Then
            expression = Mid(expression, 2): aDebug = True
            addDebug "RPN $> " & expression
        End If
    End If
    RPNcalculate = calculateRpn(convertToRpn(expression, equalityStatement), aDebug, equalityStatement)

    Exit Function
err:
    'If noerror Then RPNcalculate = "Couldn't solve expression, probably invalid.": Exit Function
    FlagError "RPN Error", err.Description & vbCrLf & "Expression: " & expression, iLocation, True
    'FlagError "Run-Time Error", err.Description, iLocation
End Function

Public Function convertToRpn(ByVal expression As String, Optional equalityStatement As Boolean = False) As String()
    'converts an expression to RPN
    
    Dim finalProduct() As String, tokenCount As Integer
    Dim tokens() As String, n As Long
    'Dim tStack As New clsStack
    Dim tStack As Integer
    tStack = addStack
    
    expression = "( " & Trim(expression) & " )"
    tokens = Tokenize(expression)
    'tokens = Split(expression, " ")
   ' MsgBox displayArray(tokens, vbCrLf)
    ReDim finalProduct(UBound(tokens))
        
    ClearStack tStack
    
    For n = LBound(tokens) To UBound(tokens)
        Select Case tokens(n)
            Case "(": Push tStack, "("
            Case ")"
                Do While Peek(tStack) <> "("
                    finalProduct(tokenCount) = Pop(tStack)
                    tokenCount = tokenCount + 1
                Loop
                If Peek(tStack) = "(" Then Pop tStack
            Case "+", "-", "*", "/", "\", "&", "^", "%", "<", ">", ">=", "<=", "<>", "!=", "=", "!", "&&", "||", "$", "|", "~", "++", "--"
                Do
                    If dStack(tStack).stackSize = 0 Or Peek(tStack) = "(" Or GetPval(Peek(tStack), equalityStatement) < GetPval(tokens(n), equalityStatement) Then
                        Push tStack, tokens(n)
                        Exit Do
                    Else
                        finalProduct(tokenCount) = Pop(tStack)
                        tokenCount = tokenCount + 1
                    End If
                Loop
            Case Else
                finalProduct(tokenCount) = tokens(n)
                tokenCount = tokenCount + 1
        End Select
    Next n
    
    removeStack tStack
    
    ReDim Preserve finalProduct(tokenCount - 1)
    
    'MsgBox displayArray(finalProduct, vbCrLf), , "Array"
    
    convertToRpn = finalProduct
End Function

Public Function calculateRpn(expression() As String, aDebug As Boolean, equalityStatement As Boolean) As String
    'solves an RPN expression
    
    Dim tokens() As String, n As Long
    Dim temp As String, temp2 As String
    Dim tStack As Integer
    Dim t1 As Integer, t2 As Integer
    Dim tempVar As typeVar, tempVar2 As typeVar
    
    tStack = addStack()
    
    'tokens = Split(Trim(displayArray(expression, " ")), Chr$(32))
    tokens = expression
    If aDebug Then addDebug "    $> " & displayArray(tokens, " ")
    
    'tStack.ClearStack
    ClearStack tStack
    
    For n = LBound(tokens) To UBound(tokens)
        If IsOper(tokens(n), False) And tokens(n) <> "!" Then CheckStack dStack(tStack).stackSize
begin:
        Select Case tokens(n)
            Case "+": Push tStack, Val(Pop(tStack)) + Val(Pop(tStack))
            Case "-": temp = Pop(tStack): Push tStack, Val(Pop(tStack)) - Val(temp)
            Case "*": Push tStack, Val(Pop(tStack)) * Val(Pop(tStack))
            Case "/": Push tStack, Val(Pop(tStack)) / Val(Pop(tStack))
            Case "\": temp = Pop(tStack): Push tStack, Val(Pop(tStack)) \ Val(temp)
            Case "^": Push tStack, Val(Pop(tStack)) ^ Val(Pop(tStack))
            Case "%": Push tStack, Val(Pop(tStack)) Mod Val(Pop(tStack))
            Case "++"
                Peek (tStack)
                doAssignment tradeVar, Convert(Val(tradeVar.value) + 1)
                Push tStack, Val(Pop(tStack)) + 1
            Case "--"
                Peek (tStack)
                doAssignment tradeVar, Convert(Val(tradeVar.value) - 1)
                Push tStack, Val(Pop(tStack)) - 1
            Case "&":
                temp = Pop(tStack)
                Push tStack, Pop(tStack) & temp
            Case "!": Push tStack, IIf(Val(Pop(tStack)) = 0, 1, 0)
            Case "<", ">", ">=", "<=", "<>", "!=", "==": temp = Pop(tStack): Push tStack, doEval(Pop(tStack), tokens(n), temp)
            Case "&&": Push tStack, IIf(Pop(tStack) <> 0 And Pop(tStack) <> 0, 1, 0)
            Case "||"
                'MsgBox displayArray(tokens, vbCrLf)
                Push tStack, IIf(Pop(tStack) <> 0 Or Pop(tStack) <> 0, 1, 0)
            Case "=":
                If equalityStatement Then tokens(n) = "==": GoTo begin
                Popp (tStack): tempVar2 = tradeVar  'right side
                Popp (tStack): tempVar = tradeVar   'left side (the variable)
                Push tStack, doAssignment(tempVar, tempVar2)
            Case Else
                If Left(tokens(n), 1) = Chr(34) Then 'quote
                    Push tStack, Mid(tokens(n), 2, Len(tokens(n)) - 2)
                ElseIf FindVar(tokens(n)) Or Trim(LCase(tokens(n))) = LCase(CurrentProc) Then  'var
                    t1 = GetVarIndex(tokens(n), False, t2)
                    If Trim(LCase(tokens(n))) = LCase(CurrentProc) Then
                        tempVar = ProcReturns(returnCount)
                        tempVar.flags = 1
                    Else
                        tempVar = dVars(t1)
                    End If
                    tempVar.cElement = t2
                    If t2 > -1 Then
                        If tempVar.isArray Then tempVar.value = tempVar.arrayValues(tempVar.cElement) Else: FlagError _
                            "Run-Time Error", "Variable type mismatch" & vbCrLf & tempVar.name & ":darray <-> dvar", iLocation
                    End If
                    tradeVar = tempVar
                    Pushh tStack
                ElseIf InStr(1, tokens(n), "(") <> 0 Then 'function
                    tradeVar = GetFunctions(tokens(n))
                    Pushh tStack
                Else 'k it's either a number or invalid, let someone else handle it :)
                    Push tStack, GetVar(tokens(n))
                End If
        End Select
    Next n
    
    If dStack(tStack).stackSize > 1 Then
        err.Raise 500, , "Invalid expression"
        End
    End If
    If aDebug Then addDebug "    $> " & Peek(tStack)
    calculateRpn = Pop(tStack)
End Function

Private Function CheckStack(tStackSize As Integer)
    Dim n As Long
    'before performing an operation, we must make sure the stack has enough stuff!
    'this prevents fatal subscript errors - new in 4.01
    If tStackSize < 2 Then
        'FlagError "Syntax Error", "Invalid expression.", iLocation
        'an error, but this will flag it up above
        n = 5 / 0
        End
    End If
End Function

Private Function GetPval(token As String, equality As Boolean) As Integer
    'these are VERY important to the RPN!  it determines which operations
    'to evaluate first.  the ones at the top are executed last and the ones
    'at the bottom are executed first.  it is NOT recommended that you change
    'these, unless you use some weird substandard math system
    Select Case token
        Case "="
            If equality Then GetPval = 4 Else GetPval = 1   '1 = assignment
        Case "||"
            GetPval = 2
        Case "&&"
            GetPval = 3
        Case "!=", "==", "<>"
            GetPval = 4
        Case ">", "<", ">=", "<="
            GetPval = 5
        Case "&"
            GetPval = 6
        Case "+", "-"
            GetPval = 7
        Case "*", "/", "\", "%"
            GetPval = 8
        Case "!"
            GetPval = 9
        Case "^"
            GetPval = 10
        Case "++", "--"
            GetPval = 11
        Case "(", ")"
            GetPval = 12
        Case Else
            GetPval = 0
    End Select
End Function

Private Function Tokenize(ByVal expression As String) As String()
    'tokenizes an expression into an array
    Dim tokenArray() As String, tokenCount As Integer
    Dim i As Integer, j As Integer, n As Long
    Dim sChar As String
    Dim includeNegative As Boolean
    
    ReDim tokenArray(Len(expression))
    
    For i = 1 To Len(expression)
        sChar = Mid(expression, i, 1)
        
        If IsOper(sChar, True) Then   'operator
            If i < Len(expression) Then
                'If IsOper(Mid(expression, i + 1, 1), False) Then _
                '    sChar = Mid(expression, i, 2): i = i + 1
                If isDoubleOp(Mid(expression, i, 2)) Then _
                    sChar = Mid(expression, i, 2): i = i + 1
            End If
            If sChar = "-" Then
                If tokenCount = 1 Then
                    includeNegative = True: GoTo skip
                Else
                    If IsOper(tokenArray(tokenCount - 1), True) Then _
                        includeNegative = True: GoTo skip
                End If
            End If
            tokenArray(tokenCount) = sChar
            tokenCount = tokenCount + 1
        ElseIf IsNumeric(sChar) Then    'number
            j = i + 1
            Do Until j > Len(expression)
                If Not IsNumeric(Mid(expression, j, 1)) And Mid(expression, j, 1) <> "." Then Exit Do
                j = j + 1
            Loop
            tokenArray(tokenCount) = Mid(expression, i, j - i)
            If includeNegative Then tokenArray(tokenCount) = -1 * Val(tokenArray(tokenCount)): _
                includeNegative = False
            tokenCount = tokenCount + 1
            i = j - 1
        ElseIf sChar = Chr(34) Then 'quote: fixed string
            j = InStr(i + 1, expression, Chr(34))
            If j > 0 Then
                tokenArray(tokenCount) = Mid(expression, i, j - i + 1)
                tokenCount = tokenCount + 1
                i = j
            Else
                err.Raise 500, , "Expected end-quote"
                End
            End If
            If includeNegative Then includeNegative = False
        Else        'other: variable? function?
            If Not isValidChar(sChar) Then GoTo skip
            
            j = i + 1
            Do Until j > Len(expression)
                If Not isValidChar(Mid(expression, j, 1)) Then Exit Do
                sChar = Mid(expression, j, 1)
                If sChar = "(" Or sChar = "[" Then
                    j = FindEncloseEnd(j + 1, expression, sChar)
                    'an error, but this will flag it up above
                    If j = 0 Then err.Raise 500, , "Expected end-parenthesis"
                End If
                j = j + 1
            Loop
            tokenArray(tokenCount) = Mid(expression, i, j - i)
            tokenCount = tokenCount + 1
            i = j - 1
        End If
skip:
    Next 'i
    
    ReDim Preserve tokenArray(tokenCount - 1)
    Tokenize = tokenArray
End Function

Private Function IsOper(ByVal NumVal As String, Parens As Boolean) As Boolean
    'determines if a character is an operator
    Select Case NumVal
        Case "+", "-", "*", "/", "\", "&", "^", "%", "<", ">", "=", "!", "|"
            IsOper = True
        Case "(", ")"
            If Parens Then IsOper = True
    End Select
End Function

Private Function isDoubleOp(ByVal Op As String) As Boolean
    Select Case Op
        Case "++", "--", ">=", "<=", "!=", "<>", "||", "&&", "=="
            isDoubleOp = True
    End Select
End Function

Public Function displayArray(theArray() As String, delimiter As String) As String
    Dim tempString As String, i As Integer
    
    For i = LBound(theArray) To UBound(theArray)
        tempString = tempString & theArray(i) & IIf(i <> UBound(theArray), delimiter, "")
    Next i
    
    displayArray = tempString
End Function

Private Function isValidChar(ByVal Char As String) As Boolean
    'determines if a character is valid
    Dim charCode As Integer
    charCode = Asc(Mid(Char, 1, 1))

    If charCode >= 65 And charCode <= 90 Then GoTo validChar      'upper case
    If charCode >= 97 And charCode <= 122 Then GoTo validChar     'lower case
    If charCode >= 48 And charCode <= 57 Then GoTo validChar      'numbers
    If charCode = 95 Or charCode = 46 Then GoTo validChar        '_ & .
    If charCode = 91 Or charCode = 93 Then GoTo validChar        ' [ ]
    If charCode = 40 Or charCode = 41 Then GoTo validChar        ' ( )
    'If charCode = 33 Then GoTo validChar
    
    Exit Function
    
validChar:
    isValidChar = True
End Function

'more stuff for independent training
'Public Function GetVar(varname As String) As String
'    If IsNumeric(varname) Then GetVar = varname: Exit Function
'
'    Select Case varname
'        Case "var1"
'            GetVar = "variable #1"
'        Case "var2"
'            GetVar = 238
'        Case "var3"
'            GetVar = "var: 238"
'        Case "crlf"
'            GetVar = vbCrLf
'        Case Else
'            MsgBox "undefined var: " & varname
'            End
'    End Select
'End Function

Public Function doAssignment(LeftVal As typeVar, RightVal As typeVar)
'old HandleExpression.  this version is way better
Dim tempVar As typeVar, elementNum As Integer

    elementNum = LeftVal.cElement
    'MsgBox LeftVal.name & vbCrLf & RightVal.value, , "ASSIGNMENT"
    
    If LeftVal.name = "" Then _
        FlagError "Syntax Error", "Left side of the equation must be valid identifier", iLocation: Exit Function
    
    If LeftVal.isArray Then   'see if it's an array
        If LeftVal.cElement > -1 Then GoTo normalMode
        If RightVal.isArray And RightVal.cElement = -1 Then  'make sure the right is also an array
            LeftVal.arrayValues = RightVal.arrayValues   'just copy the values
            LeftVal.numElements = RightVal.numElements   'and the # of elements
        Else
            FlagError "Run-Time Error", "Variable type mismatch" & vbCrLf & RightVal.name & ": darray <-> dvar", iLocation
        End If
    Else    'normal var; procede normally
normalMode:
        If RightVal.isArray And RightVal.cElement = -1 Then _
            FlagError "Run-Time Error", "Variable type mismatch" & vbCrLf & RightVal.name & ": dvar <-> darray", iLocation
            
        If LeftVal.isArray Then     'if it's an array we're assigning a value to an element of that array
            checkArrayBounds LeftVal.numElements, LeftVal.cElement, LeftVal.name
            LeftVal.arrayValues(LeftVal.cElement) = RightVal.value
        Else
            LeftVal.value = RightVal.value
        End If
    End If
    
    LeftVal.cElement = -1
    'now plug our variable in where it needs to be
    If LeftVal.flags = 0 Then
        dVars(LeftVal.myIndex) = LeftVal    'normal var
    Else
        ProcReturns(returnCount) = LeftVal  'return var
    End If
End Function

Public Function Convert(value As String, Optional name As String) As typeVar
Convert.value = value
If name = "" Then name = "*AZUREOWNS*"
Convert.name = name
Convert.isArray = False
End Function
